@@ -1,10 +1,25 @@
-# -*- coding: utf-8 -*-
-'''
-:author: Sogeti
-:data : 06 February 2019
-'''
+# Copyright 2019-2021 Davide Mancusi, Martin Maurey, Jonathan Faustin
+#
+# This file is part of t4_geom_convert.
+#
+# t4_geom_convert is free software: you can redistribute it and/or modify it
+# under the terms of the GNU General Public License as published by the Free
+# Software Foundation, either version 3 of the License, or (at your option) any
+# later version.
+#
+# t4_geom_convert is distributed in the hope that it will be useful, but
+# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
+# more details.
+#
+# You should have received a copy of the GNU General Public License along with
+# t4_geom_convert.  If not, see <https://www.gnu.org/licenses/>.
+#
+# vim: set fileencoding=utf-8 :
+
 from math import pi, fabs
 
+from ..Progress import Progress
 from ..Surface.ESurfaceTypeMCNP import ESurfaceTypeMCNP as MS
 from ..Surface.CollectionDict import CollectionDict
 from .ESurfaceTypeT4 import ESurfaceTypeT4 as T4S
@@ -14,31 +29,22 @@ from .SurfaceConversionError import SurfaceConversionError
 
 
 def convert_mcnp_surfaces(dic_surface_mcnp):
-    '''
-    :brief: method which convert MCNP surface and constructing the
-    dictionary of Surface T4
-    '''
+    '''Method which convert MCNP surface and constructing the dictionary of
+    Surface T4.'''
     dic_surface_t4 = CollectionDict()
 
-    free_id = max(int(k) for k in dic_surface_mcnp.keys()) + 1
-    n_surfaces = len(dic_surface_mcnp)
-    fmt_string = ('\rconverting surface {{:{0}d}} ({{:{1}d}}/{{:{1}d}}, '
-                  '{{:3d}}%)'
-                  .format(len(str(max(dic_surface_mcnp))),
-                          len(str(n_surfaces))))
-    for i, (key, val) in enumerate(dic_surface_mcnp.items()):
-        percent = int(100.0*i/(n_surfaces-1)) if n_surfaces > 1 else 100
-        print(fmt_string.format(key, i+1, n_surfaces, percent),
-              end='', flush=True)
+    with Progress('converting surface',
+                  len(dic_surface_mcnp), max(dic_surface_mcnp)) as progress:
+        for i, (key, val) in enumerate(dic_surface_mcnp.items()):
+            progress.update(i, key)
+            t4_surfs = convert_mcnp_surface(key, val)
+            dic_surface_t4[key] = t4_surfs
 
-        t4_surfs = convert_mcnp_surface(key, val)
-        dic_surface_t4[key] = t4_surfs
-
-    print('... done', flush=True)
     return dic_surface_t4
 
 
 def convert_mcnp_surface(key, val):
+    '''Perform the actual conversion of an MCNP surface.'''
     surf_colls = []
     for surf, side in val:
         try:
@@ -55,7 +61,6 @@ def convert_mcnp_surface(key, val):
     return t4_surfs
 
 
-
 def conversion_surface_params(key, val):
     '''Convert the MCNP surface described by `val` into a TRIPOLI-4 surface.'''
     if val.type_surface in (MS.K_X, MS.K_Y, MS.K_Z, MS.KX, MS.KY, MS.KZ, MS.K):
@@ -87,16 +92,16 @@ def convert_plane(val):
     tuple_param = val.param_surface
     p_x, p_y, p_z = tuple_param[0]
     u_x, u_y, u_z = tuple_param[1]
-    pos = -(u_x*p_x + u_y*p_y + u_z*p_z)
+    pos = -(u_x * p_x + u_y * p_y + u_z * p_z)
     if u_x == 0. and u_y == 0. and u_z > 0.:
         type_surface = T4S.PLANEZ
-        param = [-pos/u_z]
+        param = [-pos / u_z]
     elif u_y == 0. and u_z == 0. and u_x > 0.:
         type_surface = T4S.PLANEX
-        param = [-pos/u_x]
+        param = [-pos / u_x]
     elif u_z == 0. and u_x == 0. and u_y > 0.:
         type_surface = T4S.PLANEY
-        param = [-pos/u_y]
+        param = [-pos / u_y]
     else:
         type_surface = T4S.PLANE
         param = [u_x, u_y, u_z, pos]
@@ -148,11 +153,11 @@ def convert_special_quadric(val):
     ysq = sq_params[8]
     zsq = sq_params[9]
     gq_params = [asq, bsq, csq, 0.0, 0.0, 0.0,
-                 2.0*dsq - 2.0*asq*xsq,
-                 2.0*esq - 2.0*bsq*ysq,
-                 2.0*fsq - 2.0*csq*zsq,
-                 asq*xsq**2 + bsq*ysq**2 + csq*zsq**2
-                 - 2.0*(dsq*xsq + esq*ysq + fsq*zsq) + gsq]
+                 2.0 * dsq - 2.0 * asq * xsq,
+                 2.0 * esq - 2.0 * bsq * ysq,
+                 2.0 * fsq - 2.0 * csq * zsq,
+                 asq * xsq**2 + bsq * ysq**2 + csq * zsq**2
+                 - 2.0 * (dsq * xsq + esq * ysq + fsq * zsq) + gsq]
     if eval_quadric(gq_params, (xsq, ysq, zsq)) > 0.0:
         gq_params = [-param for param in gq_params]
     return T4S.QUAD, gq_params
@@ -173,10 +178,11 @@ def eval_quadric(params, point):
     >>> eval_quadric(quad, (2.000, 0.0, 0.0)) > 0.0
     True
     '''
-    x, y, z = point
-    return (params[0]*x**2 + params[1]*y**2 + params[2]*z**2
-            + params[3]*x*y + params[4]*y*z + params[5]*z*x
-            +params[6]*x + params[7]*y + params[8]*z + params[9])
+    x, y, z = point  # pylint: disable=invalid-name
+    return (params[0] * x**2 + params[1] * y**2 + params[2] * z**2
+            + params[3] * x * y + params[4] * y * z + params[5] * z * x
+            + params[6] * x + params[7] * y + params[8] * z + params[9])
+
 
 def convert_quadric(val):
     '''Convert the parameters for a quadric.'''
@@ -206,7 +212,7 @@ def convert_cone(key, val):
     '''Convert the parameters for cones.'''
     p_x, p_y, p_z = val.param_surface[0]
     u_x, u_y, u_z = val.param_surface[1]
-    theta = 180.*val.compl_param[1]/pi
+    theta = 180. * val.compl_param[1] / pi
     if u_x == 0. and u_y == 0.:
         type_surface = T4S.CONEZ
         param = [p_x, p_y, p_z, theta]
@@ -225,16 +231,16 @@ def convert_cone(key, val):
     if nappe is None or nappe == 0:
         return SurfaceCollection([(cone, 1)])
 
-    pos = -(u_x*p_x + u_y*p_y + u_z*p_z)
+    pos = -(u_x * p_x + u_y * p_y + u_z * p_z)
     if u_x == 0 and u_y == 0:
         type_surface = T4S.PLANEZ
-        param = [-pos/u_z]
+        param = [-pos / u_z]
     elif u_y == 0 and u_z == 0:
         type_surface = T4S.PLANEX
-        param = [-pos/u_x]
+        param = [-pos / u_x]
     elif u_z == 0 and u_x == 0:
         type_surface = T4S.PLANEY
-        param = [-pos/u_y]
+        param = [-pos / u_y]
     else:
         type_surface = T4S.PLANE
         param = [u_x, u_y, u_z, pos]
